#include <stdio.h>
#include <windows.h>
#include <setupapi.h>
#include <devguid.h>
#include <regstr.h>
#include <tchar.h>

#include "resource.h"

#include "FCDHidCmd.h"
#include "Registry.h"

static HINSTANCE _hInst=NULL;
static HWND _hDlg=NULL;
static TCHAR _szVIDPID[]=_T("Vid_04d8&Pid_fb56");

typedef struct 
{
	TCHAR *pszDesc;
	unsigned __int8 u8Val;

} COMBOITEMSTRUCT;

typedef struct
{
	unsigned __int8 u8CommandSet;
	unsigned __int8 u8CommandGet;
	int nIdxDefault;
	int nID;
	const COMBOITEMSTRUCT *pacis;
} COMBOSTRUCT;

static const COMBOITEMSTRUCT _cisLNAGain[]=
{
	{_T("-5.0dB"),TLGE_N5_0DB},
	{_T("-2.5dB"),TLGE_N2_5DB},
	{_T("+0.0dB"),TLGE_P0_0DB},
	{_T("+2.5dB"),TLGE_P2_5DB},
	{_T("+5,0dB"),TLGE_P5_0DB},
	{_T("+7.5dB"),TLGE_P7_5DB},
	{_T("+10.0dB"),TLGE_P10_0DB},
	{_T("+12.5dB"),TLGE_P12_5DB},
	{_T("+15.0dB"),TLGE_P15_0DB},
	{_T("+17.5dB"),TLGE_P17_5DB},
	{_T("+20.0dB"),TLGE_P20_0DB},
	{_T("+25.0dB"),TLGE_P25_0DB},
	{_T("+30.0dB"),TLGE_P30_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisLNAEnhance[]=
{
	{_T("Off"),TLEE_OFF},
	{_T("0"),TLEE_0},
	{_T("1"),TLEE_1},
	{_T("2"),TLEE_2},
	{_T("3"),TLEE_3},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisBand[]=
{
	{_T("VHF II"),TBE_VHF2},
	{_T("VHF III"),TBE_VHF3},
	{_T("UHF"),TBE_UHF},
	{_T("L band"),TBE_LBAND},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisRFFilter0[]=
{
	{_T("268MHz LPF"),TRFE_LPF268MHZ},
	{_T("299MHz LPF"),TRFE_LPF268MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisRFFilter1[]=
{
	{_T("509MHz LPF"),TRFE_LPF509MHZ},
	{_T("656MHz LPF"),TRFE_LPF656MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisRFFilter2[]=
{
	{_T("360MHz BPF"),TRFE_BPF360MHZ},
	{_T("380MHz BPF"),TRFE_BPF380MHZ},
	{_T("405MHz BPF"),TRFE_BPF405MHZ},
	{_T("425MHz BPF"),TRFE_BPF425MHZ},
	{_T("450MHz BPF"),TRFE_BPF450MHZ},
	{_T("475MHz BPF"),TRFE_BPF475MHZ},
	{_T("505MHz BPF"),TRFE_BPF505MHZ},
	{_T("540MHz BPF"),TRFE_BPF540MHZ},
	{_T("575MHz BPF"),TRFE_BPF575MHZ},
	{_T("615MHz BPF"),TRFE_BPF615MHZ},
	{_T("670MHz BPF"),TRFE_BPF670MHZ},
	{_T("720MHz BPF"),TRFE_BPF720MHZ},
	{_T("760MHz BPF"),TRFE_BPF760MHZ},
	{_T("840MHz BPF"),TRFE_BPF840MHZ},
	{_T("890MHz BPF"),TRFE_BPF890MHZ},
	{_T("970MHz BPF"),TRFE_BPF970MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisRFFilter3[]=
{
	{_T("1300MHz BPF"),TRFE_BPF1300MHZ},
	{_T("1320MHz BPF"),TRFE_BPF1320MHZ},
	{_T("1360MHz BPF"),TRFE_BPF1360MHZ},
	{_T("1410MHz BPF"),TRFE_BPF1410MHZ},
	{_T("1445MHz BPF"),TRFE_BPF1445MHZ},
	{_T("1460MHz BPF"),TRFE_BPF1460MHZ},
	{_T("1490MHz BPF"),TRFE_BPF1490MHZ},
	{_T("1530MHz BPF"),TRFE_BPF1530MHZ},
	{_T("1560MHz BPF"),TRFE_BPF1560MHZ},
	{_T("1590MHz BPF"),TRFE_BPF1590MHZ},
	{_T("1640MHz BPF"),TRFE_BPF1640MHZ},
	{_T("1660MHz BPF"),TRFE_BPF1660MHZ},
	{_T("1680MHz BPF"),TRFE_BPF1680MHZ},
	{_T("1700MHz BPF"),TRFE_BPF1700MHZ},
	{_T("1720MHz BPF"),TRFE_BPF1720MHZ},
	{_T("1750MHz BPF"),TRFE_BPF1750MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisMixerGain[]=
{
	{_T("4dB"),TMGE_P4_0DB},
	{_T("12dB"),TMGE_P12_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisBiasCurrent[]=
{
	{_T("00 L band"),TBCE_LBAND},
	{_T("01"),TBCE_1},
	{_T("10"),TBCE_2},
	{_T("11 V/U band"),TBCE_VUBAND},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisMixerFilter[]=
{
	{_T("1.9MHz"),TMFE_1_9MHZ},
	{_T("2.3MHz"),TMFE_2_3MHZ},
	{_T("2.7MHz"),TMFE_2_7MHZ},
	{_T("3.0MHz"),TMFE_3_0MHZ},
	{_T("3.4MHz"),TMFE_3_4MHZ},
	{_T("3.8MHz"),TMFE_3_8MHZ},
	{_T("4.2MHz"),TMFE_4_2MHZ},
	{_T("4.6MHz"),TMFE_4_6MHZ},
	{_T("27MHz"),TMFE_27_0MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGain1[]=
{
	{_T("-3dB"),TIG1E_N3_0DB},
	{_T("+6dB"),TIG1E_P6_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGainMode[]=
{
	{_T("Linearity"),TIGME_LINEARITY},
	{_T("Sensitivity"),TIGME_SENSITIVITY},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFRCFilter[]=
{
	{_T("1.0MHz"),TIRFE_1_0MHZ},
	{_T("1.2MHz"),TIRFE_1_2MHZ},
	{_T("1.8MHz"),TIRFE_1_8MHZ},
	{_T("2.6MHz"),TIRFE_2_6MHZ},
	{_T("3.4MHz"),TIRFE_3_4MHZ},
	{_T("4.4MHz"),TIRFE_4_4MHZ},
	{_T("5.3MHz"),TIRFE_5_3MHZ},
	{_T("6.4MHz"),TIRFE_6_4MHZ},
	{_T("7.7MHz"),TIRFE_7_7MHZ},
	{_T("9.0MHz"),TIRFE_9_0MHZ},
	{_T("10.6MHz"),TIRFE_10_6MHZ},
	{_T("12.4MHz"),TIRFE_12_4MHZ},
	{_T("14.7MHz"),TIRFE_14_7MHZ},
	{_T("17.6MHz"),TIRFE_17_6MHZ},
	{_T("21.0MHz"),TIRFE_21_0MHZ},
	{_T("21.4MHz"),TIRFE_21_4MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGain2[]=
{
	{_T("0dB"),TIG2E_P0_0DB},
	{_T("+3dB"),TIG2E_P3_0DB},
	{_T("+6dB"),TIG2E_P6_0DB},
	{_T("+9dB"),TIG2E_P9_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGain3[]=
{
	{_T("0dB"),TIG3E_P0_0DB},
	{_T("+3dB"),TIG3E_P3_0DB},
	{_T("+6dB"),TIG3E_P6_0DB},
	{_T("+9dB"),TIG3E_P9_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGain4[]=
{
	{_T("0dB"),TIG4E_P0_0DB},
	{_T("+1dB"),TIG4E_P1_0DB},
	{_T("+2dB"),TIG4E_P2_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFFilter[]=
{
	{_T("2.15MHz"),TIFE_2_15MHZ},
	{_T("2.20MHz"),TIFE_2_20MHZ},
	{_T("2.24MHz"),TIFE_2_24MHZ},
	{_T("2.28MHz"),TIFE_2_28MHZ},
	{_T("2.30MHz"),TIFE_2_30MHZ},
	{_T("2.40MHz"),TIFE_2_40MHZ},
	{_T("2.45MHz"),TIFE_2_45MHZ},
	{_T("2.50MHz"),TIFE_2_50MHZ},
	{_T("2.55MHz"),TIFE_2_55MHZ},
	{_T("2.60MHz"),TIFE_2_60MHZ},
	{_T("2.70MHz"),TIFE_2_70MHZ},
	{_T("2.75MHz"),TIFE_2_75MHZ},
	{_T("2.80MHz"),TIFE_2_80MHZ},
	{_T("2.90MHz"),TIFE_2_90MHZ},
	{_T("2.95MHz"),TIFE_2_95MHZ},
	{_T("3.00MHz"),TIFE_3_00MHZ},
	{_T("3.10MHz"),TIFE_3_10MHZ},
	{_T("3.20MHz"),TIFE_3_20MHZ},
	{_T("3.30MHz"),TIFE_3_30MHZ},
	{_T("3.40MHz"),TIFE_3_40MHZ},
	{_T("3.60MHz"),TIFE_3_60MHZ},
	{_T("3.70MHz"),TIFE_3_70MHZ},
	{_T("3.80MHz"),TIFE_3_80MHZ},
	{_T("3.90MHz"),TIFE_3_90MHZ},
	{_T("4.10MHz"),TIFE_4_10MHZ},
	{_T("4.30MHz"),TIFE_4_30MHZ},
	{_T("4.40MHz"),TIFE_4_40MHZ},
	{_T("4.60MHz"),TIFE_4_60MHZ},
	{_T("4.80MHz"),TIFE_4_80MHZ},
	{_T("5.00MHz"),TIFE_5_00MHZ},
	{_T("5.30MHz"),TIFE_5_30MHZ},
	{_T("5.50MHz"),TIFE_5_50MHZ},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGain5[]=
{
	{_T("+3dB"),TIG5E_P3_0DB},
	{_T("+6dB"),TIG5E_P6_0DB},
	{_T("+9dB"),TIG5E_P9_0DB},
	{_T("+12dB"),TIG5E_P12_0DB},
	{_T("+15dB"),TIG5E_P15_0DB},
	{NULL,0}
};

static const COMBOITEMSTRUCT _cisIFGain6[]=
{
	{_T("+3dB"),TIG6E_P3_0DB},
	{_T("+6dB"),TIG6E_P6_0DB},
	{_T("+9dB"),TIG6E_P9_0DB},
	{_T("+12dB"),TIG6E_P12_0DB},
	{_T("+15dB"),TIG6E_P15_0DB},
	{NULL,0}
};

#define FCD_HID_CMD_SET_LNA_GAIN     110 // Send a 1 byte value, see enums for reference
#define FCD_HID_CMD_SET_LNA_ENHANCE  111
#define FCD_HID_CMD_SET_BAND         112
#define FCD_HID_CMD_SET_RF_FILTER    113
#define FCD_HID_CMD_SET_MIXER_GAIN   114
#define FCD_HID_CMD_SET_BIAS_CURRENT 115
#define FCD_HID_CMD_SET_MIXER_FILTER 116
#define FCD_HID_CMD_SET_IF_GAIN1     117
#define FCD_HID_CMD_SET_IF_GAIN_MODE 118
#define FCD_HID_CMD_SET_IF_RC_FILTER 119
#define FCD_HID_CMD_SET_IF_GAIN2     120
#define FCD_HID_CMD_SET_IF_GAIN3     121
#define FCD_HID_CMD_SET_IF_FILTER    122
#define FCD_HID_CMD_SET_IF_GAIN4     123
#define FCD_HID_CMD_SET_IF_GAIN5     124
#define FCD_HID_CMD_SET_IF_GAIN6     125
static COMBOSTRUCT _acs[]=
{
	FCD_HID_CMD_SET_LNA_GAIN,    FCD_HID_CMD_GET_LNA_GAIN,    10,IDC_COMBOLNAGAIN,    _cisLNAGain,
	FCD_HID_CMD_SET_LNA_ENHANCE, FCD_HID_CMD_GET_LNA_ENHANCE,  0,IDC_COMBOLNAENHANCE, _cisLNAEnhance,
	FCD_HID_CMD_SET_BAND,        FCD_HID_CMD_GET_BAND,         0,IDC_COMBOBAND,       _cisBand,
	FCD_HID_CMD_SET_RF_FILTER,   FCD_HID_CMD_GET_RF_FILTER,    0,IDC_COMBORFFILTER,   _cisRFFilter1,
	FCD_HID_CMD_SET_MIXER_GAIN,  FCD_HID_CMD_GET_MIXER_GAIN,   1,IDC_COMBOMIXERGAIN,  _cisMixerGain,
	FCD_HID_CMD_SET_BIAS_CURRENT,FCD_HID_CMD_GET_BIAS_CURRENT, 3,IDC_COMBOBIASCURRENT,_cisBiasCurrent,
	FCD_HID_CMD_SET_MIXER_FILTER,FCD_HID_CMD_GET_MIXER_FILTER, 0,IDC_COMBOMIXERFILTER,_cisMixerFilter,
	FCD_HID_CMD_SET_IF_GAIN1,    FCD_HID_CMD_GET_IF_GAIN1,     1,IDC_COMBOIFGAIN1,    _cisIFGain1,
	FCD_HID_CMD_SET_IF_GAIN_MODE,FCD_HID_CMD_GET_IF_GAIN_MODE, 0,IDC_COMBOIFGAINMODE, _cisIFGainMode,
	FCD_HID_CMD_SET_IF_RC_FILTER,FCD_HID_CMD_GET_IF_RC_FILTER, 0,IDC_COMBOIFRCFILTER, _cisIFRCFilter,
	FCD_HID_CMD_SET_IF_GAIN2,    FCD_HID_CMD_GET_IF_GAIN2,     0,IDC_COMBOIFGAIN2,    _cisIFGain2,
	FCD_HID_CMD_SET_IF_GAIN3,    FCD_HID_CMD_GET_IF_GAIN3,     0,IDC_COMBOIFGAIN3,    _cisIFGain3,
	FCD_HID_CMD_SET_IF_GAIN4,    FCD_HID_CMD_GET_IF_GAIN4,     0,IDC_COMBOIFGAIN4,    _cisIFGain4,
	FCD_HID_CMD_SET_IF_FILTER,   FCD_HID_CMD_GET_IF_FILTER,    0,IDC_COMBOIFFILTER,   _cisIFFilter,
	FCD_HID_CMD_SET_IF_GAIN5,    FCD_HID_CMD_GET_IF_GAIN5,     0,IDC_COMBOIFGAIN5,    _cisIFGain5,
	FCD_HID_CMD_SET_IF_GAIN6,    FCD_HID_CMD_GET_IF_GAIN6,     0,IDC_COMBOIFGAIN6,    _cisIFGain6,
	0,0,0,-1,NULL
};

static void PopulateCombo(HWND hDlg,int nID,int nIdxDefault,const COMBOITEMSTRUCT *pcis)
{
	SendDlgItemMessage(hDlg,nID,CB_RESETCONTENT,0,0);

	while (pcis->pszDesc!=NULL)
	{
		SendDlgItemMessage(hDlg,nID,CB_ADDSTRING,0,(LPARAM)pcis->pszDesc);
		pcis++;
	}
	SendDlgItemMessage(hDlg,nID,CB_SETCURSEL,(WPARAM)nIdxDefault,0);
}

static void PopulateCombos(HWND hDlg)
{
	COMBOSTRUCT *pcs=_acs;

	while (pcs->pacis!=NULL)
	{
		const COMBOITEMSTRUCT *pcis=pcs->pacis;

		PopulateCombo(hDlg,pcs->nID,pcs->nIdxDefault,pcis);

		pcs++;
	}
}

static BOOL HIDOpen(PTCHAR pszDevSearchID,PTCHAR pszDevID,PHANDLE phWrite,PHANDLE phRead)
{
	GUID InterfaceClassGuid = {0x4d1e55b2, 0xf16f, 0x11cf, 0x88, 0xcb, 0x00, 0x11, 0x11, 0x00, 0x00, 0x30}; 
	HDEVINFO hDevInfo;
	SP_DEVINFO_DATA DeviceInfoData;
	DWORD i;
	PTCHAR pszDevSearchIDLower=_tcsdup(pszDevSearchID);

	_tcslwr(pszDevSearchIDLower);

	// Create a HDEVINFO with all present devices.
	hDevInfo = SetupDiGetClassDevs
	(
		&InterfaceClassGuid,
		0, // Enumerator
		0,
		DIGCF_PRESENT | DIGCF_DEVICEINTERFACE 
	);
   
	if (hDevInfo == INVALID_HANDLE_VALUE)
	{
		// Insert error handling here.
		return FALSE;
	}
   
	// Enumerate through all devices in Set.
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
	for (i=0;SetupDiEnumDeviceInfo(hDevInfo,i,&DeviceInfoData);i++)
	{
		DWORD DataT;
		LPTSTR buffer = NULL;
		DWORD buffersize = 0;

		while
		(
			!SetupDiGetDeviceRegistryProperty
			(
				hDevInfo,
				&DeviceInfoData,
				SPDRP_HARDWAREID,
				&DataT,
				(PBYTE)buffer,
				buffersize,
				&buffersize
			)
		)
		{
			if (GetLastError() == ERROR_INSUFFICIENT_BUFFER)
			{
				// Change the buffer size.
				if (buffer) 
				{
					LocalFree(buffer);
				}
				// Double the size to avoid problems on 
				// W2k MBCS systems per KB 888609. 
				buffer=LocalAlloc(LPTR,buffersize * 2);
			}
			else
			{
				// Insert error handling here.
				break;
			}
		}

		_tcslwr(buffer);
       
		if (_tcsstr(buffer,pszDevSearchIDLower)!=NULL)
		{
			// We have found the device, now lets try to open a handles to it

			SP_DEVICE_INTERFACE_DATA dids;

			dids.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
			if(!SetupDiEnumDeviceInterfaces(hDevInfo, NULL, &InterfaceClassGuid, i, &dids))
			{
				break;
			}

			{
				SP_DEVICE_INTERFACE_DETAIL_DATA didds;
				PSP_DEVICE_INTERFACE_DETAIL_DATA pdidds=&didds;
				DWORD cb;

				pdidds->cbSize=sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
				SetupDiGetDeviceInterfaceDetail(hDevInfo,&dids,NULL,0,&cb,NULL);	
				if ((pdidds=malloc(cb))==NULL)
				{
					break; // uh-oh
				}
				pdidds->cbSize=sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
				SetupDiGetDeviceInterfaceDetail(hDevInfo,&dids,pdidds,cb,NULL,NULL);

				*phWrite=CreateFile(pdidds->DevicePath,GENERIC_WRITE,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,0);
				if (GetLastError()!=ERROR_SUCCESS)
				{
					break;
				}
				*phRead=CreateFile(pdidds->DevicePath,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,0);
				if (GetLastError()!=ERROR_SUCCESS)
				{
					break;
				}
			}
			SetupDiDestroyDeviceInfoList(hDevInfo);
			free(pszDevSearchIDLower);
			_tcscpy(pszDevID,buffer);
			return TRUE;
		}
       
		if (buffer)
		{
			LocalFree(buffer);
		}
	}
	if ( GetLastError()!=NO_ERROR && GetLastError()!=ERROR_NO_MORE_ITEMS )
	{
	   return FALSE;
	}
	SetupDiDestroyDeviceInfoList(hDevInfo);
	free(pszDevSearchIDLower);
	return FALSE;
}

static void HIDClose(HANDLE hWrite,HANDLE hRead)
{
	CloseHandle(hWrite);
	CloseHandle(hRead);
}

static unsigned char HIDSetFreq(HANDLE hWrite,HANDLE hRead,int nFreq)
{
	DWORD dwBytesWritten=0;
	DWORD dwBytesRead=0;
	unsigned char au8BufOut[65]; // endpoint size + 1
	unsigned char au8BufIn[65]; // endpoint size + 1

	au8BufOut[0]=0; // First byte is report ID. Ignored by HID Class firmware as only config'd for one report
	au8BufOut[1]=100;
	au8BufOut[2]=(unsigned char)nFreq;
	au8BufOut[3]=(unsigned char)(nFreq>>8);
	au8BufOut[4]=(unsigned char)(nFreq>>16);
	WriteFile(hWrite,au8BufOut,65,&dwBytesWritten,0);
	ReadFile(hRead,au8BufIn,65,&dwBytesRead,0);
	return au8BufIn[2];
}

static unsigned char HIDResetFCD(HANDLE hWrite,HANDLE hRead)
{
	DWORD dwBytesWritten=0;
	DWORD dwBytesRead=0;
	unsigned char au8BufOut[65]; // endpoint size + 1
	unsigned char au8BufIn[65]; // endpoint size + 1

	au8BufOut[0]=0; // First byte is report ID. Ignored by HID Class firmware as only config'd for one report
	au8BufOut[1]=255;
	WriteFile(hWrite,au8BufOut,65,&dwBytesWritten,0);
	ReadFile(hRead,au8BufIn,65,&dwBytesRead,0);
	return au8BufIn[2];
}

static void Message(PTCHAR psz)
{
	SendMessage(GetDlgItem(_hDlg,IDC_EDITMESSAGE),EM_REPLACESEL,(WPARAM)FALSE,(LPARAM)psz);
}

static void MessageClear(void)
{
	if (_hDlg!=NULL)
	{
		SetDlgItemText(_hDlg,IDC_EDITMESSAGE,_T(""));
	}
}

static BOOL HIDReadCommand(HANDLE hWrite,HANDLE hRead,unsigned __int8 u8Cmd,unsigned __int8 *pu8Data,unsigned __int8 u8Len)
{
	DWORD dwBytesWritten=0;
	DWORD dwBytesRead=0;
	unsigned char au8BufOut[65]; // endpoint size + 1
	unsigned char au8BufIn[65]; // endpoint size + 1

	// Clean out any outstanding reads
	
	au8BufOut[0]=0; // First byte is report ID. Ignored by HID Class firmware as only config'd for one report
	au8BufOut[1]=u8Cmd;
	Sleep(10);
	WriteFile(hWrite,au8BufOut,65,&dwBytesWritten,0);
	ReadFile(hRead,au8BufIn,65,&dwBytesRead,0);
	memcpy(pu8Data,au8BufIn+3,u8Len);
	return au8BufIn[2];
}

static BOOL HIDQuery(HANDLE hWrite,HANDLE hRead,unsigned __int8 *pu8Data,unsigned __int8 u8Len)
{
	DWORD dwBytesWritten=0;
	DWORD dwBytesRead=0;
	unsigned char au8BufOut[65]; // endpoint size + 1
	unsigned char au8BufIn[65]; // endpoint size + 1

	// Clean out any outstanding reads
	
	au8BufOut[0]=0; // First byte is report ID. Ignored by HID Class firmware as only config'd for one report
	au8BufOut[1]=FCD_HID_CMD_QUERY;
	Sleep(10);
	WriteFile(hWrite,au8BufOut,65,&dwBytesWritten,0);
	ReadFile(hRead,au8BufIn,65,&dwBytesRead,0);
	memcpy(pu8Data,au8BufIn+3,u8Len);
	return au8BufIn[2];
}

static BOOL CheckVersion(HANDLE hWrite,HANDLE hRead,char *pszVersion)
{
	char szVersionRead[60];

	if (HIDQuery(hWrite,hRead,szVersionRead,60)!=TRUE)
	{
		if (strncmp(szVersionRead,"FCDAPP",6)==0)
		{

		}
		else
		{
			Message(_T("\r\nError, FCD not in application mode."));
		}
	}
	else
	{
		Message(_T("\r\nError, FCD firmware too low, Query command not supported."));
	}
}

static void ReadCommand(unsigned __int8 u8Cmd,unsigned __int8 *pu8Data,unsigned __int8 u8Len)
{
	TCHAR sz[200];
	TCHAR szMsg[400];
	BOOL b;
	HANDLE hWrite=INVALID_HANDLE_VALUE;
	HANDLE hRead=INVALID_HANDLE_VALUE;

	MessageClear();
	b=HIDOpen(_szVIDPID,sz,&hWrite,&hRead);
	_stprintf(szMsg,"HIDOpen of %s returns %d",_szVIDPID,b);
	Message(szMsg);
	if (b)
	{
		unsigned __int8 u8;

		Message(_T("\r\nSuccess, FCD found."));
		u8=HIDReadCommand(hWrite,hRead,u8Cmd,pu8Data,u8Len);
		_stprintf(szMsg,"\r\nHIDReadCommand returns %d",u8);
		Message(szMsg);
		if (u8==1)
		{
			_stprintf(szMsg,"\r\nSuccess, FCD read command, value: %d",*pu8Data);
			Message(szMsg);
		}
		else
		{
			Message(_T("\r\nError, FCD read command."));
		}
	}
	else
	{
		Message(_T("\r\nError, FCD not found: is FCD correctly installed?"));
	}
	HIDClose(hWrite,hRead);
}

static BOOL HIDWriteCommand(HANDLE hWrite,HANDLE hRead,unsigned __int8 u8Cmd,unsigned __int8 *pu8Data,unsigned __int8 u8Len)
{
	DWORD dwBytesWritten=0;
	DWORD dwBytesRead=0;
	unsigned char au8BufOut[65]; // endpoint size + 1
	unsigned char au8BufIn[65]; // endpoint size + 1

	// Clean out any outstanding reads
	au8BufOut[0]=0; // First byte is report ID. Ignored by HID Class firmware as only config'd for one report
	au8BufOut[1]=u8Cmd;
	memcpy(au8BufOut+2,pu8Data,u8Len);
	Sleep(10);
	WriteFile(hWrite,au8BufOut,65,&dwBytesWritten,0);
	ReadFile(hRead,au8BufIn,65,&dwBytesRead,0);
	return au8BufIn[2];
}

static void WriteCommand(unsigned __int8 u8Cmd,unsigned __int8 *pu8Data,unsigned __int8 u8Len)
{
	TCHAR sz[200];
	TCHAR szMsg[400];
	BOOL b;
	HANDLE hWrite=INVALID_HANDLE_VALUE;
	HANDLE hRead=INVALID_HANDLE_VALUE;

	MessageClear();
	b=HIDOpen(_szVIDPID,sz,&hWrite,&hRead);
	_stprintf(szMsg,"HIDOpen of %s returns %d",_szVIDPID,b);
	Message(szMsg);
	if (b)
	{
		unsigned __int8 u8;

		Message(_T("\r\nSuccess, FCD found."));
		u8=HIDWriteCommand(hWrite,hRead,u8Cmd,pu8Data,u8Len);
		_stprintf(szMsg,"\r\nHIDWriteCommand returns %d",u8);
		Message(szMsg);
		if (u8==1)
		{
			Message("\r\nSuccess, FCD write register.");
		}
		else
		{
			Message(_T("\r\nError, FCD read register."));
		}
	}
	else
	{
		Message(_T("\r\nError, FCD not found: is FCD correctly installed?"));
	}
	HIDClose(hWrite,hRead);
}

static void BandChange(HWND hDlg)
{
	static const COMBOITEMSTRUCT *apcis[4]={_cisRFFilter0,_cisRFFilter1,_cisRFFilter2,_cisRFFilter3};
	int nIdx=SendDlgItemMessage(hDlg,IDC_COMBOBAND,CB_GETCURSEL,0,0);
	int nIdxOrg=SendDlgItemMessage(hDlg,IDC_COMBORFFILTER,CB_GETCURSEL,0,0);

	PopulateCombo(hDlg,IDC_COMBORFFILTER,nIdxOrg,apcis[nIdx]);
}

static void ReadDevice(HWND hDlg)
{
	COMBOSTRUCT *pcs=_acs;
//	unsigned __int16 au16[50];
	int i=0;

	while (pcs->pacis!=NULL)
	{
		unsigned __int8 u8;
		const COMBOITEMSTRUCT *pcis=pcs->pacis;
		int nIdx=0;

		ReadCommand(pcs->u8CommandGet,&u8,1);

		// Try to find the index to the register field value
		while (pcis!=NULL && pcis->u8Val!=u8)
		{
			nIdx++;
			pcis++;
		}

		if (pcis!=NULL)
		{
			SendDlgItemMessage(hDlg,pcs->nID,CB_SETCURSEL,(WPARAM)nIdx,0);
		}
		else
		{
			SendDlgItemMessage(hDlg,pcs->nID,CB_SETCURSEL,(WPARAM)-1,0); // clear the combo box if cannot find
		}
		pcs++;
		i++;
	}
	BandChange(hDlg);
}

static void SetFreq(int nFreq)
{
	TCHAR sz[200];
	TCHAR szMsg[400];
	BOOL b;
	HANDLE hWrite=INVALID_HANDLE_VALUE;
	HANDLE hRead=INVALID_HANDLE_VALUE;

	MessageClear();
	b=HIDOpen(_szVIDPID,sz,&hWrite,&hRead);
	_stprintf(szMsg,"HIDOpen of %s returns %d",_szVIDPID,b);
	Message(szMsg);
	if (b)
	{
		unsigned int u8;

		Message(_T("\r\nSuccess, FCD found."));
		u8=HIDSetFreq(hWrite,hRead,nFreq);
		_stprintf(szMsg,"\r\nHIDSetFreq of %d returns %d",nFreq,u8);
		Message(szMsg);
		if (u8==1)
		{
			ReadDevice(_hDlg);
			Message(_T("\r\nSuccess, FCD frequency set."));
		}
		else
		{
			Message(_T("\r\nError, FCD frequency NOT set."));
		}
	}
	else
	{
		Message(_T("\r\nError, FCD not found: is FCD correctly installed?"));
	}
	HIDClose(hWrite,hRead);
}

static void ResetFCD(void)
{
	TCHAR sz[200];
	TCHAR szMsg[400];
	BOOL b;
	HANDLE hWrite=INVALID_HANDLE_VALUE;
	HANDLE hRead=INVALID_HANDLE_VALUE;

	MessageClear();
	b=HIDOpen(_szVIDPID,sz,&hWrite,&hRead);
	_stprintf(szMsg,"HIDOpen of %s returns %d",_szVIDPID,b);
	Message(szMsg);
	if (b)
	{
		unsigned int u8;

		Message(_T("\r\nSuccess, FCD found."));
		u8=HIDResetFCD(hWrite,hRead);
		_stprintf(szMsg,"\r\nHIDReset returns %d",u8);
		Message(szMsg);
		if (u8==1)
		{
			Message(_T("\r\nSuccess, FCD reset to bootloader."));
		}
		else
		{
			Message(_T("\r\nError, FCD NOT reset."));
		}
	}
	else
	{
		Message(_T("\r\nError, FCD not found: is FCD correctly installed?"));
	}
	HIDClose(hWrite,hRead);
}

static void WriteParameter(HWND hDlg,int nIdx) // nIdx is index into _acs of COMBOSTRUCT
{
	COMBOSTRUCT *pcs=_acs+nIdx;
	int nIdxSel=SendDlgItemMessage(hDlg,pcs->nID,CB_GETCURSEL,0,0);

	if (nIdxSel!=-1)
	{
		unsigned __int8 u8Write=pcs->pacis[nIdxSel].u8Val;

		WriteCommand(pcs->u8CommandSet,&u8Write,1);
	}
}

static void WriteDevice(HWND hDlg)
{
	COMBOSTRUCT *pcs=_acs;
	int nIdx=0;

	while (pcs->pacis!=NULL)
	{
		WriteParameter(hDlg,nIdx);
		pcs++;
		nIdx++;
	}
}

static void SetDefaults(HWND hDlg)
{
	COMBOSTRUCT *pcs=_acs;

	while (pcs->pacis!=NULL)
	{
		SendDlgItemMessage(hDlg,pcs->nID,CB_SETCURSEL,(WPARAM)pcs->nIdxDefault,0);
		pcs++;
	}
}

static LRESULT CALLBACK MainDlg(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam)
{
	static HFONT hFont=INVALID_HANDLE_VALUE;
	static HBRUSH hBrush;
	static const COLORREF crEditFreqBk=RGB(0,0,0);
	static const COLORREF crEditFreq=RGB(0,255,0);
	static HBITMAP hBmp=INVALID_HANDLE_VALUE;

	switch (message) 
	{
		case WM_INITDIALOG:
			_hDlg=hDlg;
			hBrush = CreateSolidBrush(crEditFreqBk);
		  hFont=CreateFont (48, 0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS, _T("Arial"));
			SendDlgItemMessage(hDlg,IDC_EDITFREQ,WM_SETFONT,(WPARAM)hFont,TRUE);
			SendDlgItemMessage(hDlg,IDC_STATICKHZ,WM_SETFONT,(WPARAM)hFont,TRUE);
			hBmp=LoadBitmap(_hInst,MAKEINTRESOURCE(IDB_BITMAP3)); 
			SendDlgItemMessage(hDlg,IDC_STATICPIC,STM_SETIMAGE,(WPARAM)IMAGE_BITMAP,(LPARAM)hBmp);
			{
				TCHAR sz[30]=_T("999885");

				RegGetString("FreqCorr",sz,30-1);
				SetDlgItemText(hDlg,IDC_EDITFREQCORR,sz);
			}

			{
				SCROLLINFO si;

				si.fMask=SIF_RANGE | SIF_PAGE | SIF_POS;
				si.cbSize=sizeof(si);
				si.nMax=32767;
				si.nMin=-32768;
				si.nPage=256;
				si.nPos=0;

				{
					TCHAR sz[30]=_T("0");

					RegGetString("DCICorr",sz,30-1);
					si.nPos=_ttoi(sz);
					_stprintf(sz,_T("%.5f"),si.nPos/32768.0);
					SetDlgItemText(hDlg,IDC_EDITDCI,sz);
				}
				SetScrollInfo(GetDlgItem(hDlg,IDC_SCROLLBARFINEDCI),SB_CTL,&si,TRUE);
				{
					TCHAR sz[30]=_T("0");

					RegGetString("DCQCorr",sz,30-1);
					si.nPos=_ttoi(sz);
					_stprintf(sz,_T("%.5f"),si.nPos/32768.0);
					SetDlgItemText(hDlg,IDC_EDITDCQ,sz);
				}
				SetScrollInfo(GetDlgItem(hDlg,IDC_SCROLLBARFINEDCQ),SB_CTL,&si,TRUE);
				{
					TCHAR sz[30]=_T("0");

					RegGetString("PhaseCorr",sz,30-1);
					si.nPos=_ttoi(sz);
					_stprintf(sz,_T("%.5f"),si.nPos/32768.0);
					SetDlgItemText(hDlg,IDC_EDITPHASE,sz);
				}
				SetScrollInfo(GetDlgItem(hDlg,IDC_SCROLLBARFINEPHASE),SB_CTL,&si,TRUE);

				si.fMask=SIF_RANGE | SIF_PAGE | SIF_POS;
				si.cbSize=sizeof(si);
				si.nMax=65535;
				si.nMin=0;
				si.nPage=256;
				{
					TCHAR sz[30]=_T("32768");

					RegGetString("GainCorr",sz,30-1);
					si.nPos=_ttoi(sz);
					_stprintf(sz,_T("%.5f"),si.nPos/32768.0);
					SetDlgItemText(hDlg,IDC_EDITGAIN,sz);
				}
				SetScrollInfo(GetDlgItem(hDlg,IDC_SCROLLBARFINEGAIN),SB_CTL,&si,TRUE);
			}
//			SendDlgItemMessage(hDlg,IDC_COMBOLNAGAIN,CB_ADDSTRING,0,(LPARAM)_T("ABC"));

			PopulateCombos(hDlg);			
			ReadDevice(hDlg);
			SetTimer(hDlg,1,1000,NULL);
			break;
		case WM_DESTROY:
			KillTimer(hDlg,1);
			DeleteObject(hBmp);
			hBmp=INVALID_HANDLE_VALUE;
			DeleteObject(hFont);
			hFont=INVALID_HANDLE_VALUE;
			DeleteObject(hBrush);
			hBrush=INVALID_HANDLE_VALUE;
			_hDlg=NULL;
      return TRUE;
		case WM_TIMER:
			{
				unsigned char u8=0;
				TCHAR sz[30];

				ReadCommand(FCD_HID_CMD_GET_PLL_LOCK,&u8,1);
				CheckDlgButton(hDlg,IDC_CHECKPLL,u8==1 ? BST_CHECKED : BST_UNCHECKED);
				u8=0;
				ReadCommand(FCD_HID_CMD_GET_IF_RSSI,&u8,1);
				_itot(u8,sz,10);
				SetDlgItemText(hDlg,IDC_EDITRSSI,sz);
			}
			break;
		case WM_CTLCOLOREDIT:
			if ((HWND)lParam==GetDlgItem(hDlg,IDC_EDITFREQ))
			{
				SetTextColor((HDC)wParam,crEditFreq);
				SetBkColor((HDC)wParam,crEditFreqBk);
				return (LRESULT)hBrush;
			}
			break;
		case WM_VSCROLL:
			{
				SCROLLINFO si;
				int nLineInc=1;

				si.cbSize=sizeof(si);
				si.fMask=SIF_ALL;
				GetScrollInfo((HWND)lParam,SB_CTL,&si);

				if ((GetAsyncKeyState(VK_CONTROL) & 0x8000)!=0)
				{
					nLineInc=si.nPage;
				}

        switch (LOWORD(wParam))
        {
					case SB_TOP:
						si.nPos=si.nMin;
            break;
		      case SB_LINEUP:
						si.nPos+=nLineInc;
	          break;
		      case SB_PAGEUP:
						si.nPos+=si.nPage;
	          break;
					case SB_LINEDOWN:
						si.nPos-=nLineInc;
						break;
					case SB_PAGEDOWN:
						si.nPos-=si.nPage;
						break;
	        case SB_THUMBPOSITION:
            si.nPos=HIWORD(wParam);
            break;
					case SB_THUMBTRACK:
						si.nPos=HIWORD(wParam);
						break;
					case SB_BOTTOM:
						si.nPos=si.nMax;
						break;
					case SB_ENDSCROLL:
						break;
				}
				if (si.nPos>si.nMax)
				{
					si.nPos=si.nMax;
				}
				if (si.nPos<si.nMin)
				{
					si.nPos=si.nMin;
				}

        SetScrollPos((HWND)lParam,SB_CTL,si.nPos,FALSE);

				{
					double d=((double)si.nPos)/32768.0;
					TCHAR sz[30];

					_stprintf(sz,_T("%.5f"),d);

					switch (GetDlgCtrlID((HWND)lParam))
					{
						case IDC_SCROLLBARFINEDCI:
							{
								TCHAR sz[30];

								_itot(si.nPos,sz,10);
								RegSetString(_T("DCICorr"),sz);
							}
							SetDlgItemText(hDlg,IDC_EDITDCI,sz);
							break;
						case IDC_SCROLLBARFINEDCQ:
							{
								TCHAR sz[30];

								_itot(si.nPos,sz,10);
								RegSetString(_T("DCQCorr"),sz);
							}
							SetDlgItemText(hDlg,IDC_EDITDCQ,sz);
							break;
						case IDC_SCROLLBARFINEGAIN:
							{
								TCHAR sz[30];

								_itot(si.nPos,sz,10);
								RegSetString(_T("GainCorr"),sz);
							}
							SetDlgItemText(hDlg,IDC_EDITGAIN,sz);
							break;
						case IDC_SCROLLBARFINEPHASE:
							{
								TCHAR sz[30];

								_itot(si.nPos,sz,10);
								RegSetString(_T("PhaseCorr"),sz);
							}
							SetDlgItemText(hDlg,IDC_EDITPHASE,sz);
							break;
					}
				}

				{
					unsigned char auc[4];
					int n;

					n=GetScrollPos(GetDlgItem(hDlg,IDC_SCROLLBARFINEDCI),SB_CTL);
					memcpy(auc+0,&n,2);
					n=GetScrollPos(GetDlgItem(hDlg,IDC_SCROLLBARFINEDCQ),SB_CTL);
					memcpy(auc+2,&n,2);
					WriteCommand(FCD_HID_CMD_SET_DC_CORR,auc,4);

					n=GetScrollPos(GetDlgItem(hDlg,IDC_SCROLLBARFINEPHASE),SB_CTL);
					memcpy(auc+0,&n,2);
					n=GetScrollPos(GetDlgItem(hDlg,IDC_SCROLLBARFINEGAIN),SB_CTL);
					memcpy(auc+2,&n,2);
					WriteCommand(FCD_HID_CMD_SET_IQ_CORR,auc,4);
				}
			}
			break;
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDOK:
				case IDCANCEL:
				case ID_EXIT:
					EndDialog(hDlg, TRUE);
					break;
				case ID_SETFREQ:
					{
						int n;
						int nCorr;
						TCHAR sz[30];
						TCHAR szCorr[30];

						GetDlgItemText(hDlg,IDC_EDITFREQ,sz,sizeof(sz)-1);
						GetDlgItemText(hDlg,IDC_EDITFREQCORR,szCorr,sizeof(szCorr)-1);
						n=_ttoi(sz);
						nCorr=_ttoi(szCorr);
						n=(int)(n*(nCorr/1000000.0));

						if (n<50000 || n>2500000)
						{
							MessageBox(hDlg,_T("Target frequency must be between 50000 and 2500000"),_T("Error"),MB_OK | MB_ICONEXCLAMATION);
						}
						else
						{
							SetFreq(n);
						}
					}
					break;

				case IDC_EDITFREQCORR:
					switch (HIWORD(wParam))
					{
						case EN_CHANGE:
							{
								TCHAR sz[30];

								GetDlgItemText(hDlg,IDC_EDITFREQCORR,sz,30-1);
								RegSetString(_T("FreqCorr"),sz);
							}
							break;
						default:
							break;
					}
					break;

					break;
				case ID_READDEVICE:
					ReadDevice(hDlg);
					break;
				case ID_WRITEDEVICE:
					WriteDevice(hDlg);
					ReadDevice(hDlg);
					break;
				case ID_DEFAULTS:
					SetDefaults(hDlg);
					WriteDevice(hDlg);
					ReadDevice(hDlg);
					break;

/*
				case IDC_COMBOBAND:
					switch (HIWORD(wParam))
					{
						case CBN_SELCHANGE:
							BandChange(hDlg);
							break;
						default:
							break;
					} // No break here!
*/
				case IDC_COMBOLNAGAIN:
				case IDC_COMBOLNAENHANCE:
				case IDC_COMBOBAND:
				case IDC_COMBORFFILTER:
				case IDC_COMBOMIXERGAIN:
				case IDC_COMBOBIASCURRENT:
				case IDC_COMBOMIXERFILTER:
				case IDC_COMBOIFGAIN1:
				case IDC_COMBOIFGAINMODE:
				case IDC_COMBOIFRCFILTER:
				case IDC_COMBOIFGAIN2:
				case IDC_COMBOIFGAIN3:
				case IDC_COMBOIFGAIN4:
				case IDC_COMBOIFFILTER:
				case IDC_COMBOIFGAIN5:
				case IDC_COMBOIFGAIN6:
					switch (HIWORD(wParam))
					{
						case CBN_SELCHANGE:
							{
								COMBOSTRUCT *pcs=_acs;
								int nIdx=-1; // Index into _acs array of combo box item

								while (pcs->pacis!=NULL && nIdx==-1)
								{
									if (pcs->nID==LOWORD(wParam))
									{
										nIdx=pcs-_acs;
									}
									else
									{
										pcs++;
									}
								}
								if (nIdx!=-1)
								{
									WriteParameter(hDlg,nIdx);
									ReadDevice(hDlg);
								}
							}
							break;
						default:
							break;
					}
					break;

				case ID_RESETFCD:
					ResetFCD();
					break;
				default:
					break;
			}
      return TRUE;

		default:
			break;
	}
  return FALSE;
}

int APIENTRY WinMain(HINSTANCE hInst,HINSTANCE hPrevInst,LPSTR lpCmdLine,int nCmdShow)
{
	_hInst=hInst;
	DialogBox(hInst,MAKEINTRESOURCE(IDD_MAIN),NULL,MainDlg);
  return 0;
}
